// For licensing, see: http://hydrogen.knownspace.org/license.html
//
// $Id: Pool.java,v 1.44 2002/10/14 19:05:53 cjdent Exp $
// 
// Integrated persistence layer with Pool.  This is more of a future
// enhancement due to the facts that 
// 1.  the Pool is a Singleton
// 2.  The persistence of Pool already occurs at the EntityManager level.  
// 3.  A Pool is useless without an EntityManager's
// Identifier/Entity/EntityProxy mappings.
//
// Updated javadocs to reflect new integration with persistence. -- baa 
//
// Rewrote javadocs for EventGeneratorAssistant member variable.
//
// Moved responsibility for null checking from fireEvent and unsubscribe
// to loadAssistant().  Renamed loadNewAssistant() to loadAssistant()
// Made the Pool's default constructor private and added a createDefaultPool()
// method that returns the DefaultPool.  If multiple pools become a necessity,
// a PoolManager class is recommended for usage (which the createDefaultPool() 
// method will need to register with).  
// search() method:
//
// Additional concerns/errata:  
// How might we go about ordering search results (instead of binary
// satisfied/unsatisfied constraints, what about fuzzy logic-style  
// constraints, i.e., search matched 50%, 47%, 37%, 23%).  Degree of
// acceptance?  Is this an issue? -- BAA
///////////////////////////////////////////////////////////////////////////////
// 2002-10-13 cam/cjd
// 1. Made Serializable.
// 2. Turned on serializable style of persistence, as that works.
///////////////////////////////////////////////////////////////////////////////
package org.datamanager.kernel;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OptionalDataException;
import java.io.Serializable;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.datamanager.constraint.EntityConstraint;
import org.datamanager.constraint.EventConstraint;

import org.datamanager.event.AddEntityPoolEvent;
import org.datamanager.event.DataManagerEvent;
import org.datamanager.event.SearchPoolEvent;
import org.datamanager.event.SubscribeEventGeneratorEvent;
import org.datamanager.event.UnsubscribeEventGeneratorEvent;

import org.datamanager.exception.DataManagerException;

import org.datamanager.kernel.EntityManager;
import org.datamanager.kernel.Persistable;
import org.datamanager.kernel.VerySimpleEntityManager;

import org.datamanager.persistence.DataAccessObject;
import org.datamanager.persistence.DataAccessObjectFactory;

import org.datamanager.tools.Debug;

/**
 * The Pool is a collection of Entities. It can be
 * searched, subscribed to, and modified, by adding
 * and removing Entities.
 *
 * @author Matthew Farrellee
 * @version $Revision: 1.44 $
 */
public class Pool
    implements EventGenerator, Serializable
    {
    /** String represention of the default persisted location of the Pool. */
    public final static String PERSISTENCE_DEFAULT_FILE_LOCATION = 
            Persistable.PERSISTENCE_FOLDER + File.separator + 
            DataManager.DEFAULT_POOL + ".xml";
    /** File handle for the default persisted location of the Pool. */
    final static File PERSISTENCE_DEFAULT_FILE = new File(PERSISTENCE_DEFAULT_FILE_LOCATION);
    /** Pool is currently a singleton. */
    private static Pool instance;
    /**
     * The EventGeneratorAssistant helps the EventGenerator implementation of
     * Pool handle its subscribe, and unsubscribe requests.  It also manages
     * the fireEvent request.  It installs new EventHandlers, removes extant 
     * EventHandlers, and propagates events generated by the pool or passed in 
     * via the add(DataManagerEvent event) method.
     */
    private transient EventGeneratorAssistant assistant;
    /**
     * The collection of Entities itself. Currently not very sophisticated. 
     */
    private List pool;

    /**
     * The Pool's default constructor. (uninstantiatable)
     */
    private Pool()
        {
        assistant = new EventGeneratorAssistant();
        pool = new ArrayList();
        }

    /**
     * The EventGeneratorAssistant is transient, so this
     * method is used to reload the EventGeneratorAssistant
     * when needed.
     */
    protected void loadAssistant()
        {
        if (assistant == null)
            {
            assistant = new EventGeneratorAssistant();
            }
        }

    /**
     * Fires an event. This method is used internally in the
     * Pool for firing events generated by the Pool.
     *
     * @param event An event to fire.
     */
    protected void fireEvent(DataManagerEvent event)
        {
        loadAssistant();
        assistant.fireEvent(event);
        }

    /**
     * Reports the default Pool.
     *
     * @return The only instance of the Pool.
     *
     * @throws DataManagerException
     */
    public static Pool getDefaultPool()
                               throws DataManagerException
        {
        return EntityManager.getInstance().getPool();
        }

    /**
     * Searches for all entities in the Pool satisfying the parameterized
     * constraint. In the future this may be better served by returning an
     * immutable enumeration/iterator over the constrained entities.
     *
     *
     * @param constraint A constraint to filter entities from the Pool.
     *
     * @return An array of Entities that satisfied the constraint.
     *
     * @throws DataManagerException
     */
    public synchronized Entity[] search(EntityConstraint constraint)
                                 throws DataManagerException
        {
        fireEvent(new SearchPoolEvent((EventGenerator) this, constraint));

        Entity entity;
        EntityManager proxyManager = EntityManager.getInstance();

        List acceptedEntities = new ArrayList();
        Iterator entities = pool.iterator();

        while (entities.hasNext())
            {
            entity = proxyManager.findProxy((Identifier) entities.next());

            if (constraint.accepts(entity))
                {
                acceptedEntities.add(entity);
                }
            }

        Entity[] searchResults = new Entity[acceptedEntities.size()];

        return (Entity[]) acceptedEntities.toArray(searchResults);
        }

    /** 
     * Creates the default Pool.  If there is to be but a single Pool at any
     * point in time, this should return the lone instance of the pool in lieu
     * of generating new pools each time.  If there are to be multiple pools,
     * there should be some sort of Pool Manager (currently EntityManager)
     * that should be notified as to the creation of yet another pool. (and
     * this class should delegate to that manager).  If there is an
     * XML-persisted version of this Pool, it restores that version of this
     * Pool via the persister's restore() method.
     *
     * Package-private (kernel level) access only.
     * @param none
     */
    static final Pool createDefaultPool()
        {
        if (instance == null)
            {
            loadDefaultPool();

            // sanity check to ensure that loadDefaultPool is doing its job.
            assert instance != null;
            }

        return instance;
        }

    private static void loadDefaultPool()
        {
        /* This stuff is not necessary right now.  We need to figure out how
        * to avoid filename collisions eventually as right now requests to
        * persist the Pool persist the EntityManager tied to the Pool at the
        * Pool's default persisted location.  Another time.
                                
        if (PERSISTENCE_DEFAULT_FILE.exists()) {
            // try to restore the Pool -- ideally, restore should have a more
            // elegant way of signaling the callee as to success/failure of
            // restoration besides just returning null.
            try {
                instance = (Pool) Persistable.PERSISTER.restore(
                                   PERSISTENCE_DEFAULT_FILE);
            } catch (Exception e) {
                Debug.writeLogMessage("Unable to restore Pool from " + 
                                      PERSISTENCE_DEFAULT_FILE + ": " + 
                                      e.getMessage());
                e.printStackTrace();
            }
        }
                                
                                
        // if the Pool instance is still null, instantiate a new one. 
        if (instance == null) {
            instance = new Pool();
        }
        */
        instance = new Pool();
        }

    /**
     * Links an entity to the Pool.
     *
     * @param entity The Entity to add to the Pool.
     *
     * @throws DataManagerException
     */
    public synchronized void add(Entity entity)
                          throws DataManagerException
        {
        pool.add(EntityManager.getInstance()
                              .findIdentifier((EntityProxy) entity));

        fireEvent(new AddEntityPoolEvent((EventGenerator) this, entity));
        }

    /**
     * Adds an event to the Pool. This event will also be
     * fired to all EventHandlers subscribed to the Pool who
     * are interested in it.
     * Note: DataManagerEvents are NOT stored as Entities are.
     *
     * @param event The event to add to the Pool.
     *
     * @throws DataManagerException
     */
    public synchronized void add(DataManagerEvent event)
                          throws DataManagerException
        {
        fireEvent(event);
        }

    /**
     * Removes all the entity's entity links as well as its
     * link to the pool.
     * Note: This is currently not implemented.
     *
     * @param entity The Entity to remove from the Pool.
     *
     * @throws DataManagerException
     */
    public synchronized void remove(Entity entity)
                             throws DataManagerException
        {
        }

    /**
     * Subscribes an EventHandler with an EntityConstraint to
     * the Pool.
     *
     * @param handler The EventHandler to subscribe.
     * @param constrain The constraint with which to filter events.
     */
    public synchronized void subscribe(EventHandler handler, 
                                       EventConstraint constraint)
                                throws DataManagerException
        {
        // fireEvent ensures that the assistant is not null by invoking
        // loadAssistant() within its method body -- this ensures that
        // assistant is not null for the below addEventHandler() call.
        fireEvent(new SubscribeEventGeneratorEvent((EventGenerator) this, 
                                                   handler, constraint));

        assistant.addEventHandler(handler, constraint);
        }

    /**
     * Unsubscribes an EventHandler.
     *
     * @param handler The EventHandler to unsubscribe.
     *
     * @throws DataManagerException
     */
    public synchronized void unsubscribe(EventHandler handler)
                                  throws DataManagerException
        {
        loadAssistant();

        assistant.removeEventHandler(handler);

        fireEvent(new UnsubscribeEventGeneratorEvent((EventGenerator) this, 
                                                     handler));
        }

    /**
     * Persists this Pool and all its Entities.  Package-private access.  
     *
     * Delegates to the VerySimpleEntityManagers persist() method so as to
     * preserve the Entity mappings necessary for querying the Pool (the Pool
     * merely maintains a List of Identifiers currently, which are useless
     * without their associated Entities/EntityProxies).
     * 
     * @param File outputfile to be persisted to.  
     */
    public void persist(File outputFile)
                 throws DataManagerException
        {
        EntityManager.getInstance().persist(outputFile);
        }

    /* This is assuming that Entities have overridden object equals() */
    public boolean equals(Object object)
        {
        boolean isEqual = (object instanceof Pool);

        if (isEqual)
            {
            try
                {
                Entity[] entities = ((Pool) object).search(
                                            EntityConstraint.NONE);
                isEqual = (entities.length == pool.size());

                for (int i = 0; (i < entities.length) && isEqual; i++)
                    {
                    isEqual = pool.contains(EntityManager.getInstance()
                                                         .findIdentifier((EntityProxy) entities[i]));
                    }
                }
            catch (Exception exception)
                {
                isEqual = false;
                exception.printStackTrace();
                }
            }

        return isEqual;
        }
    }