// For licensing, see: http://hydrogen.knownspace.org/license.html
///////////////////////////////////////////////////////////////////////////////
// 2002-10-11 baa
// Tidied javadocs/fixed some spelling mistakes.
///////////////////////////////////////////////////////////////////////////////
// 2002-10-03 cam
// No longer has direct knowledge of the KernelWindow.  Instead, we export a
// SimpletonList which the KernelWindow registers as an Observer of.
///////////////////////////////////////////////////////////////////////////////
// 2002-10-05 cam
// Removed unneeded KernelWindow import line.
// Corrected the doc comment above load().
// Improved the class doc comment.
///////////////////////////////////////////////////////////////////////////////
// 2002-10-05 cam
// Extracted configureDocumentBuilderFactory() method.
// Corrected or refined javadoc comments.
///////////////////////////////////////////////////////////////////////////////
// 2002-10-07 cam
// Changed Debug.assert() calls to Debug.writeLogMessage() to enable
// java 1.4 assertions elsewhere.
///////////////////////////////////////////////////////////////////////////////
package org.datamanager.kernel;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.datamanager.kernel.Simpleton;
import org.datamanager.kernel.SimpletonList;

import org.datamanager.tools.Debug;
import org.datamanager.tools.InformationParser;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xml.sax.SAXException;

/**
 * Looks in a given XML file and loads and instantiates all Simpletons found in
 * the file.  The Simpletons are then added to the SimpletonList created by
 * this SimpletonLoaderProcess.
 *
 * @author Matthew Farrellee
 */
class SimpletonLoaderProcess
    extends Simpleton
    {
    private static final String className = "org.datamanager.kernel.SimpletonLoaderProcess";
    private static final String name = "SimpletonLoaderProcess";
    private static final String shortDescription = "Loads Simpleton class files.";
    private static final String longDescription = "Loads Simpleton class files from a directory.";
    private static final String version = "Hydrogen";
    private static final String[] authors = {"Matthew Farrellee"};
    /** Simpleton.Information about this LoaderProcess. */
    private static final Simpleton.Information information = 
            new Simpleton.Information(className, name, shortDescription, 
                                      longDescription, version, authors);
    /** File to read StartupData from. */
    private File startupDataFile;
    /** Lists of the simpletons and interfaces we load. */
    private SimpletonList simpletonList;

    /**
     * Creates a SimpletonLoaderProcess given a File that points
     * to a StartupData XML file.  Currently Kernel access only.
     *
     * @param directory A File that should be an XML StartupData file.
     */
    SimpletonLoaderProcess(File startupDataFile)
        {
        this.startupDataFile = startupDataFile;

        setInformation(information);

        simpletonList = new SimpletonList();
        }

    /**
     * Retrieves the list of Simpletons.  Currently Kernel access only.  
     */
    SimpletonList getSimpletonList()
        {
        return simpletonList;
        }

    /**
     * Starts the SimpletonLoaderProcess, which parses the
     * StartupData file.  Extracts all the Simpletons from the
     * StartupData file, tries to create an instance of each of those
     * Simpletons, and then calls the load() method on each new instance via 
     * the parse(Document document) method.
     * @see #parse(Document document)
     */
    public void process()
        {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        configureDocumentBuilderFactory(factory);

        try
            {
            // Create the DOM parser from the factory.
            // Can throw: ParserConfigurationException
            DocumentBuilder builder = factory.newDocumentBuilder();

            // Parse the XML file into a DOM tree.
            // Can throw: SAXException, IOException
            Document document = builder.parse(startupDataFile);

            // Parse the DOM tree.
            parse(document);
            }
        catch (ParserConfigurationException configurationException)
            {
            configurationException.printStackTrace();
            }
        catch (SAXException saxException)
            {
            saxException.printStackTrace();
            }
        catch (IOException ioException)
            {
            ioException.printStackTrace();
            }
        }

    /**
     * Configures the document builder factory with the required settings.
     */
    private void configureDocumentBuilderFactory(DocumentBuilderFactory factory)
        {
        // Enable support for XML namespaces.
        factory.setNamespaceAware(true);

        // Enable validation for XML documents.
        factory.setValidating(true);

        factory.setIgnoringComments(true);

        // Enable elimination of whitespace in element content.
        // FIXME: This does not work!!!
        factory.setIgnoringElementContentWhitespace(true);

        // Enable conversion of CDATA nodes to Text nodes appending
        // to adjacent text nodes.
        factory.setCoalescing(true);

        factory.setExpandEntityReferences(true);
        }

    /**
     * Parses the Document representing the configuration file. Creates the
     * Simpleton and UserInterfaces specified by the file.
     *
     * @param document The Document representation of the configuration file.  
     */
    protected void parse(Document document)
        {
        Element root = document.getDocumentElement();

        NodeList simpletonList = root.getElementsByTagName("Simpleton");
        NodeList userInterfaceList = root.getElementsByTagName("Userinterface");

        for (int index = 0; index < simpletonList.getLength(); index++)
            {
            parseSimpleton(simpletonList.item(index));
            }

        for (int index = 0; index < userInterfaceList.getLength(); index++)
            {
            parseSimpleton(userInterfaceList.item(index));
            }
        }

    /**
     * Parses a Node in the configuration file representing a Simpleton.
     * Instantiates the Simpleton.
     * 
     * @param simpleton Node within the XML configuration file representin' a
     * Simpleton 
     */
    protected void parseSimpleton(Node simpleton)
        {
        Class simpletonClass;
        Object loadedObject;
        Simpleton loadedSimpleton;

        Simpleton.Information information = InformationParser.parseInformation(
                                                    simpleton);

        // If the information is incomplete parseInformation returns NONE
        if (information != Information.NONE)
            {
            try
                {
                simpletonClass = Class.forName(information.getClassName());

                if (!simpletonClass.isInterface())
                    {
                    loadedObject = simpletonClass.newInstance();

                    loadedSimpleton = (Simpleton) loadedObject;

                    loadedSimpleton.setInformation(information);

                    load(loadedSimpleton);
                    }
                }
            catch (ClassNotFoundException classNotFoundException)
                {
                // This will only happen if Class.forName() fails.
                Debug.writeLogMessage("Cannot load: " + 
                                      information.getClassName());
                classNotFoundException.printStackTrace();
                }
            catch (IllegalAccessException noAccess)
                {
                // This should hopefully never happen.
                Debug.writeLogMessage("Cannot access: " + 
                                      information.getClassName());
                noAccess.printStackTrace();
                }
            catch (InstantiationException createError)
                {
                // For some reason an newInstance did not work. It is
                // possible that there is not a zero argument
                // constructor in the Class.
                Debug.writeLogMessage("Cannot create an instance of '" + 
                                      information.getClassName() + "'");
                }
            catch (ClassCastException wrongObjectType)
                {
                // The loadedObject was not of the right type.
                // Ignoring the class...
                Debug.writeLogMessage(information.getClassName() + 
                                      " is of wrong type: " + 
                                      wrongObjectType);
                }
            catch (SecurityException securityException)
                {
                securityException.printStackTrace();
                }
            }
        }

    /**
     * Reports Simpleton.Information about this LoaderProcess.
     *
     * @return Simpleton.Information about this LoaderProcess.
     */
    public Simpleton.Information getInformation()
        {
        return information;
        }

    /**
     * Loads the given Simpleton into the SimpletonList.
     *
     * @param simpleton A Simpleton to load.
     */
    protected void load(Simpleton simpleton)
        {
        simpletonList.add(simpleton);
        }
    }