// For licensing, see: http://hydrogen.knownspace.org/license.html

/**
 * NumberEntityValue.java
 *
 * Created 199911080021
 *
 * @author Benjamin Keil
 */
package org.datamanager.passiveentityvalue;

/**
 * NumberEntityValue is an abstract parent class for Double- and
 * IntegerEntityValues.
 */

// Table of Contents
// -----------------
// o-o Comparisons - compare with numberic primitives and types, and other
// : :   NumericEntityValues
// : :
// : :-+ Comparable Interface Methods - for convenience and sorting purposes,
// : :     the NumericEntityValue interface extends Comparable, which gives one
// : :     method that indicates greater, lesser, or equal.  I find this
// : :     distasteful, as using its result requires switching or ifing
// : :     around.  For that reason, there are a host of other comparison
// : :     methods available.
// : :
// : :-o Ranges
// : : :-o boolean isInRange(NumericEntityValue low, NumericEntityValue high)
// : : :-o boolean isInRange(Number, Number)
// : : `-o boolean isInRange(double low, double high)
// : :
// : :-o Greater-thans
// : : :-o boolean isGreaterThan(NumericEntityValue low)
// : : :-o boolean isGreaterThan(Number low)
// : : :-o boolean isGreaterThan(double low)
// : : :-o boolean isGreaterThanOrEqualTo(NumericEntityValue low)
// : : :-o boolean isGreaterThanOrEqualTo(Number low)
// : : `-o boolean isGreaterThanOrEqualTo(double low)
// : :
// : :-+ Less-thans (as above)
// : `-o Equality Tests (and HashCode stuff)
// :   :-o boolean equals(double)
// :   :-o boolean equals(Object) (should behave sanely with
// :   :     NumericEntityValue and Number Objects)
// :   `-o hashCode()
// :
// :-o Conversions
// : :
// : :-o To int
// : : :-o int getNearestInt() -- rounds to even integer when two integers are
// : : :     equidistant.
// : : :-o int getFloor() -- greatest integer less than the EntityValue
// : : `-o int getCeiling() -- least integer greater than the EntityValue
// : :
// : `-o To double
// :   `-o double getNearestDouble()
// :
// `-o Arithmetic
//   :-o Addition -- stores the result in the augend (this)
//   : :-o void add(NumericEntityValue addend)
//   : :-o void add(Number addend)
//   : `-o void add(double addend)
//   :
//   :-+ Subtraction (as above, except: subtract(...), result stored in
//   :     minuend)
//   :-+ Multiplication -- multiplyBy(...)
//   `-+ Division -- divideBy(...)
public abstract class NumberEntityValue
    implements NumericEntityValue
    {
    ///////////////////////////////////////////////////////////////////////////
    //                                                                       //
    //                              Comparisons                              //
    //                                                                       //
    ///////////////////////////////////////////////////////////////////////////
    // ------------------- Comparable Interface Methods -----------------------
    public int compareTo(NumericEntityValue n)
        {
        if (this.isLesserThan(n))
            {
            return -1;
            }
        else if (this.equals(n))
            {
            return 0;
            }
        else if (this.isGreaterThan(n))
            {
            return 1;
            }
        else
            {
            throw new Error("Fatal error in " + "IntegerEntityValue." + 
                            "compareTo(NumericEntityValue)");
            }
        }

    public int compareTo(double d)
        {
        if (this.isLesserThan(d))
            {
            return -1;
            }
        else if (this.equals(d))
            {
            return 0;
            }
        else if (this.isGreaterThan(d))
            {
            return 1;
            }
        else
            {
            throw new Error("Fatal error in " + "IntegerEntityValue." + 
                            "compareTo(double)");
            }
        }

    // That this code must be written is a sad side-effect of Java's two-faced
    // attitude towards polymorphism...
    public int compareTo(Object o)
        {
        if (o instanceof NumericEntityValue)
            {
            return this.compareTo((NumericEntityValue) o);
            }
        else if (o instanceof Number)
            {
            return this.compareTo(((Number) o).doubleValue());
            }
        else
            {
            throw new ClassCastException("Uncomparable object in " + 
                                         "IntegerEntityValue" + 
                                         ".compareTo(Object)");
            }
        }

    // -----------------------        Ranges       ----------------------------
    public boolean isInRange(NumericEntityValue low, NumericEntityValue high)
        {
        return (low.getNearestDouble() <= getValue()) && 
               (high.getNearestDouble() >= getValue());
        }

    public boolean isInRange(double low, double high)
        {
        return (low <= getValue()) && (high >= getValue());
        }

    // -----------------------    Greater Thans    ----------------------------
    public boolean isGreaterThan(NumericEntityValue low)
        {
        return getValue() > low.getNearestDouble();
        }

    public boolean isGreaterThan(Number low)
        {
        return getValue() > low.doubleValue();
        }

    public boolean isGreaterThan(double low)
        {
        return getValue() > low;
        }

    ///////////////////////////////////////////////////////////////////////////
    public boolean isGreaterThanOrEqualTo(NumericEntityValue low)
        {
        return getValue() >= low.getNearestDouble();
        }

    public boolean isGreaterThanOrEqualTo(Number low)
        {
        return getValue() >= low.doubleValue();
        }

    public boolean isGreaterThanOrEqualTo(double low)
        {
        return getValue() >= low;
        }

    // -----------------------    Lesser  Thans    ----------------------------
    public boolean isLesserThan(NumericEntityValue high)
        {
        return getValue() < high.getNearestDouble();
        }

    public boolean isLesserThan(Number high)
        {
        return getValue() < high.doubleValue();
        }

    public boolean isLesserThan(double high)
        {
        return getValue() < high;
        }

    ///////////////////////////////////////////////////////////////////////////
    public boolean isLesserThanOrEqualTo(NumericEntityValue high)
        {
        return getValue() <= high.getNearestDouble();
        }

    public boolean isLesserThanOrEqualTo(Number high)
        {
        return getValue() <= high.doubleValue();
        }

    public boolean isLesserThanOrEqualTo(double high)
        {
        return getValue() <= high;
        }

    // -----------------------   Equality Tests    ----------------------------
    public boolean equals(NumericEntityValue n)
        {
        return n.getNearestDouble() == getValue();
        }

    public boolean equals(double d)
        {
        return d == getValue();
        }

    public boolean equals(Object o)
        {
        if (o instanceof NumericEntityValue)
            {
            return ((NumericEntityValue) o).getNearestDouble() == getValue();
            }
        else if (o instanceof Number)
            {
            return ((Number) o).doubleValue() == getValue();
            }
        else
            {
            return false;
            }
        }

    public abstract int hashCode();

    ///////////////////////////////////////////////////////////////////////////
    //                                                                       //
    //                             Conversions                               //
    //                                                                       //
    ///////////////////////////////////////////////////////////////////////////
    // ----------------------     To Integer      -----------------------------
    public abstract int getNearestInt();

    public abstract int getFloor();

    public abstract int getCeiling();

    // ----------------------     To Double      ------------------------------
    public abstract double getNearestDouble();

    // ----------------------      To String     ------------------------------
    public abstract String toString();

    //////////////////////////////////////////////////////////////////////////
    //                                                                      //
    //                              Arithmetic                              //
    //                                                                      //
    //////////////////////////////////////////////////////////////////////////
    public void add(NumericEntityValue addend)
        {
        add(addend.getNearestDouble());
        }

    public void add(Number addend)
        {
        add(addend.doubleValue());
        }

    public void add(double addend)
        {
        setValue(getValue() + addend);
        }

    ///////////////////////////////////////////////////////////////////////////
    public void subtract(NumericEntityValue subtrahend)
        {
        subtract(subtrahend.getNearestDouble());
        }

    public void subtract(Number subtrahend)
        {
        subtract(subtrahend.doubleValue());
        }

    public void subtract(double subtrahend)
        {
        setValue(getValue() - subtrahend);
        }

    ///////////////////////////////////////////////////////////////////////////
    public void multiplyBy(NumericEntityValue multiplier)
        {
        multiplyBy(multiplier.getNearestDouble());
        }

    public void multiplyBy(Number multiplier)
        {
        multiplyBy(multiplier.doubleValue());
        }

    public void multiplyBy(double multiplier)
        {
        setValue(getValue() * multiplier);
        }

    ///////////////////////////////////////////////////////////////////////////
    public void divideBy(NumericEntityValue divisor)
        {
        divideBy(divisor.getNearestDouble());
        }

    public void divideBy(Number divisor)
        {
        divideBy(divisor.doubleValue());
        }

    public void divideBy(double divisor)
        {
        setValue(getValue() / divisor);
        }

    //////////////////////////////////////////////////////////////////////////
    //                                                                      //
    // These methods allow the various numeric types to share a lot of code //
    //                                                                      //
    //////////////////////////////////////////////////////////////////////////
    abstract double getValue();

    abstract void setValue(double value);
    } // NumberEntityValue
