// For licensing, see: http://hydrogen.knownspace.org/license.html
///////////////////////////////////////////////////////////////////////////////
// 2002-10-13 mml
// Modified buildArray() so that it can now properly handle null arrays.
///////////////////////////////////////////////////////////////////////////////
package org.datamanager.persistence.xml;

import java.lang.reflect.*;

import org.datamanager.persistence.xml.fieldtypes.*;

import org.w3c.dom.*;

/**
 * write object to xml document
 * It will use reflection to get the object state and keep them xml file.
 *
 * @author xi rao & Jason Lynn Baumgartner
 */
public class XMLDataPersistence
    {
    private ObjectIDMap objectIDMap;
    private TypeMap fieldTypeMap;

    /**
     * Constructor: create XMLDataPersistence with objectIDMap and fileTypeMap
     */
    protected XMLDataPersistence()
        {
        objectIDMap = new ObjectIDMap();
        fieldTypeMap = new TypeMap();
        }

    // return ObjectIDMap instance of XMLDataPersistence
    private ObjectIDMap getObjectIDMap()
        {
        return objectIDMap;
        }

    // return FieldTypeMap instence of XMLDataPersistence
    private TypeMap getFieldTypeMap()
        {
        return fieldTypeMap;
        }

    /**
     * return a document to keep object's state
     * @param object instance
     * @return
     */
    protected Document getXMLRepresentation(Object object)
        {
        Document document = XMLDocument.getXMLDocument();
        buildHeader(object, document);
        getObjectIDMap().clear();

        return document;
        }

    // build some header information in the XML docuement used to keep object's status
    private void buildHeader(Object object, Document document)
        {
        Element javaElement = document.createElement(XMLDataAccessObject.JAVA);
        javaElement.setAttribute(XMLDataAccessObject.VERSION, 
                                 XMLDataAccessObject.JAVA_VERSION);
        document.appendChild(javaElement);
        buildObject(object, javaElement);
        }

    // build the object reference in the XML document and registed in the ObjectIDMap if it is not
    // registed
    private void buildObject(Object object, Element parent)
        {
        if (getObjectIDMap().containsObject(object))
            {
            buildObjectReference(object, parent);
            }
        else
            {
            if (object == null)
                {
                /** @todo what should we do here when object is null */
                return;
                }

            getObjectIDMap().add(object);
            buildObjectInstance(object, parent);
            }
        }

    // build the object instance in the XML docuement with idference etc.
    private void buildObjectInstance(Object object, Element parent)
        {
        Document ownerDocument = parent.getOwnerDocument();
        Element objectElement = ownerDocument.createElement(
                                        XMLDataAccessObject.OBJECT);
        String idreference = getObjectIDMap().getIDReference(object);
        objectElement.setAttribute(XMLDataAccessObject.ID, idreference);
        objectElement.setAttribute(XMLDataAccessObject.CLASS, 
                                   object.getClass().getName());
        parent.appendChild(objectElement);
        buildProperties(object.getClass(), object, objectElement);
        }

    // build the object instance with reference in the XML docuemnt
    private void buildObjectReference(Object object, Element parent)
        {
        Document ownerDocument = parent.getOwnerDocument();
        Element objectElement = ownerDocument.createElement(
                                        XMLDataAccessObject.OBJECT);
        String idreference = getObjectIDMap().getIDReference(object);
        objectElement.setAttribute(XMLDataAccessObject.IDREF, idreference);
        parent.appendChild(objectElement);
        }

    // build the peoperties, which is the state of the object in the XML document
    private void buildProperties(Class cls, Object object, Element parent)
        {
        Document ownerDocument = parent.getOwnerDocument();
        Field[] fields = cls.getDeclaredFields();

        for (int i = 0; i < fields.length; i++)
            {
            Field field = fields[i];
            field.setAccessible(true);

            Element propertyElement = ownerDocument.createElement(
                                              XMLDataAccessObject.VOID);
            propertyElement.setAttribute(XMLDataAccessObject.PROPERTY, 
                                         field.getName());
            parent.appendChild(propertyElement);

            if (field.getType().isPrimitive())
                {
                buildPrimitive(field.getType().getName(), field, object, 
                               propertyElement);
                }
            else if (field.getType().getName()
                          .equalsIgnoreCase(XMLDataAccessObject.STRING_PACKAGE))
                {
                buildPrimitive(XMLDataAccessObject.STRING, field, object, 
                               propertyElement);
                }
            else if (field.getType().isArray())
                {
                buildArray(field, object, propertyElement);
                }
            else
                {
                try
                    {
                    buildObject(field.get(object), propertyElement);
                    }
                catch (IllegalAccessException iae)
                    {
                    }
                }
            }

        Class base = cls.getSuperclass(); //object.getClass().getSuperclass();

        if (base == null)
            {
            return;
            }

        buildProperties(base, object, parent);
        }

    // build the properties ( primitive ) in the xml document
    private void buildPrimitive(String type, Field field, Object object, 
                                Element parent)
        {
        Document ownerDocument = parent.getOwnerDocument();

        try
            {
            Object objectValue = field.get(object);

            if (objectValue == null)
                {
                Element nullElement = ownerDocument.createElement(
                                              XMLDataAccessObject.NULL);
                parent.appendChild(nullElement);
                }
            else
                {
                Element primitiveElement = ownerDocument.createElement(type);
                Text text = ownerDocument.createTextNode(String.valueOf(
                                                                 objectValue));
                primitiveElement.appendChild(text);
                parent.appendChild(primitiveElement);
                }
            }
        catch (IllegalAccessException iae)
            {
            }
        }

    // build the properties (array) in the xml docuemnt
    private void buildArray(Field field, Object object, Element parent)
        {
        try
            {
            Document ownerDocument = parent.getOwnerDocument();

            if (null == field.get(object))
                {
                Element nullElement = ownerDocument.createElement(
                                              XMLDataAccessObject.NULL);
                parent.appendChild(nullElement);
                }
            else
                {
                Element arrayElement = ownerDocument.createElement(
                                               XMLDataAccessObject.ARRAY);
                parent.appendChild(arrayElement);

                Class componentType = field.getType().getComponentType();

                if (getFieldTypeMap().containsField(componentType.getName()))
                    {
                    FieldType fieldType = getFieldTypeMap()
                                              .getFieldType(componentType.getName());
                    fieldType.persistPrimitiveArray(field, object, arrayElement);
                    }
                else
                    {
                    Object[] array = (Object[]) field.get(object);

                    for (int i = 0; (array != null) && (i < array.length); i++)
                        {
                        if (array[i] instanceof Object)
                            {
                            Class type = array[i].getClass();

                            if (type.isPrimitive())
                                {
                                buildPrimitive(type.getName(), field, object, 
                                               arrayElement); // not working :-(
                                }
                            else if (type.getName()
                                         .equalsIgnoreCase(XMLDataAccessObject.STRING_PACKAGE))
                                {
                                Element stringElement = 
                                        ownerDocument.createElement(
                                                XMLDataAccessObject.STRING);
                                Text text = ownerDocument.createTextNode(
                                                    String.valueOf(array[i]));
                                stringElement.appendChild(text);
                                arrayElement.appendChild(stringElement);
                                }
                            else
                                {
                                buildObject(array[i], arrayElement);
                                }
                            }
                        }
                    }
                }
            }
        catch (IllegalAccessException iae)
            {
            }
        catch (ClassCastException cce)
            {
            }
        }
    }