///////////////////////////////////////////////////////////////////////////////
// 2002-10-10 cam
// 1. Added for testEntityProxyRemoveBase() and testEntityProxyAddBase() to
//    ensure that those methods on EntityProxy that take an EntityProxy argument
//    are properly exercised.
// 2. Corrected assertion report string in testRemoveBase.
// 3. Wrote testSubscribe() and testUnsubscribe().
///////////////////////////////////////////////////////////////////////////////
package org.datamanager.kernel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.datamanager.constraint.EntityConstraint;
import org.datamanager.constraint.EventConstraint;

import org.datamanager.event.DataManagerEvent;
import org.datamanager.event.SubscribeEventGeneratorEvent;

import org.datamanager.exception.DataManagerException;

import org.datamanager.kernel.Entity;
import org.datamanager.kernel.VerySimpleEntity;

import org.datamanager.passiveentityvalue.BooleanEntityValue;
import org.datamanager.passiveentityvalue.NullEntityValue;

/**
 * This class was created to test Entity and its collaborators.  Specifically,
 * this class tests Entity, EntityManager, VerySimpleEntity, VerySimpleEntityManager,
 * EntityProxy, Identifier, IdentifierManager, and EntityValue.  The currently pathetic
 * thing about this class is that it does not test subcribe() and unsubscribe() (see below).
 *
 * @author Mary Clegg
 * @author Ailish Byrne
 */
public class EntityTester
    extends TestCase
    {
    String testString;
    Entity eventTestEntity;
    boolean hasSubscribed;

    private EntityTester(String s)
        {
        super(s);
        }

    public static Test suite()
        {
        TestSuite suite = new TestSuite("EntityTester");
        suite.addTest(new EntityTester("testCreate"));
        suite.addTest(new EntityTester("testAttach"));
        suite.addTest(new EntityTester("testDetach"));
        suite.addTest(new EntityTester("testAddBase"));
        suite.addTest(new EntityTester("testEntityProxyAddBase"));
        suite.addTest(new EntityTester("testRemoveBase"));
        suite.addTest(new EntityTester("testEntityProxyRemoveBase"));
        suite.addTest(new EntityTester("testGetAttributes"));
        suite.addTest(new EntityTester("testGetBases"));
        suite.addTest(new EntityTester("testGetAndSetValue"));
        suite.addTest(new EntityTester("testIdentifierManager"));
        suite.addTest(new EntityTester("testSubscribe"));
        suite.addTest(new EntityTester("testUnsubscribe"));

        return suite;
        }

    public void setUp()
        {
        testString = "";
        hasSubscribed = false;
        }

    public void testCreate()
        {
        try
            {
            String entityName = "TestEntity";
            Entity testEntity = VerySimpleEntity.create(entityName);

            Assert.assertNotNull("TestEntity should not be null!", testEntity);
            Assert.assertEquals("name should equal to " + entityName + "!", 
                                entityName, testEntity.getName());
            Assert.assertNotNull("attributes should not be null!", 
                                 testEntity.getAttributes(EntityConstraint.NONE));
            Assert.assertEquals("attributes should not have any Entitys!", 0, 
                                testEntity.getAttributes(EntityConstraint.NONE).length);
            Assert.assertNotNull("bases should not be null!", 
                                 testEntity.getBases(EntityConstraint.NONE));
            Assert.assertEquals("bases should not have any Entitys!", 0, 
                                testEntity.getBases(EntityConstraint.NONE).length);
            Assert.assertEquals("value should equal NullEntityValue!", 
                                NullEntityValue.getInstance(), 
                                testEntity.getValue());
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testAttach()
        {
        UnsupportedOperationException unsupportedOperationException = null;
        DataManagerException dataManagerException = null;
        Entity testEntity = null;
        Entity attributeEntity = null;
        Entity testAttachEntity = null;
        Entity proxiedTestEntity = null;
        Entity baseEntity = null;
        List attributes = null;

        try
            {
            //set up to test attach()
            testEntity = VerySimpleEntity.create("TestEntity");
            attributeEntity = VerySimpleEntity.create("AttributeEntity");
            testAttachEntity = VerySimpleEntity.create("TestAttachEntity");
            proxiedTestEntity = ((EntityProxy) testEntity).getProxiedEntity();
            baseEntity = VerySimpleEntity.create("BaseEntity");

            testEntity.attach(attributeEntity);
            proxiedTestEntity.attach(
                    ((EntityProxy) testAttachEntity).getIdentifier());

            Entity[] attributesArray = testEntity.getAttributes(
                                               EntityConstraint.NONE);
            attributes = new ArrayList(attributesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < attributesArray.length;
                 arrayIndex++)
                {
                attributes.add(attributesArray[arrayIndex]);
                }

            proxiedTestEntity.addBase(
                    ((EntityProxy) baseEntity).getIdentifier());
            proxiedTestEntity.attach(((EntityProxy) baseEntity).getIdentifier());
            }
        catch (org.datamanager.exception.DataManagerException dataManager)
            {
            dataManagerException = dataManager;
            }
        catch (UnsupportedOperationException unsupported)
            {
            unsupportedOperationException = unsupported;
            }

        //assert tests
        Assert.assertNull(
                "UnsupportedOperationException should not have been thrown!", 
                unsupportedOperationException);
        Assert.assertTrue(
                "testEntity's attributes should have contained attributeEntity!", 
                attributes.contains(attributeEntity));
        Assert.assertTrue(
                "testEntity's attributes should have contained testAttachEntity!", 
                attributes.contains(testAttachEntity));
        Assert.assertNotNull("DataManagerException should have been thrown!", 
                             dataManagerException);
        }

    public void testDetach()
        {
        UnsupportedOperationException unsupportedOperationException = null;

        try
            {
            //set up to test detach()
            Entity testEntity = VerySimpleEntity.create("TestEntity");
            Entity attributeEntity = VerySimpleEntity.create("AttributeEntity");
            Entity testAttachEntity = VerySimpleEntity.create(
                                              "TestAttachEntity");
            Entity proxiedTestEntity = ((EntityProxy) testEntity).getProxiedEntity();

            testEntity.attach(attributeEntity);
            proxiedTestEntity.attach(
                    ((EntityProxy) testAttachEntity).getIdentifier());

            testEntity.detach(attributeEntity);
            proxiedTestEntity.detach(
                    ((EntityProxy) testAttachEntity).getIdentifier());

            Entity[] attributesArray = testEntity.getAttributes(
                                               EntityConstraint.NONE);
            List attributes = new ArrayList(attributesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < attributesArray.length;
                 arrayIndex++)
                {
                attributes.add(attributesArray[arrayIndex]);
                }

            //assert tests
            Assert.assertNull(
                    "UnsupportedOperationException should not have been thrown!", 
                    unsupportedOperationException);
            Assert.assertFalse(
                    "testEntity's attributes should not have contained attributeEntity!", 
                    attributes.contains(attributeEntity));
            Assert.assertFalse(
                    "testEntity's attributes should not have contained testAttachEntity!", 
                    attributes.contains(testAttachEntity));
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        catch (UnsupportedOperationException unsupported)
            {
            unsupportedOperationException = unsupported;
            }
        }

    public void testAddBase()
        {
        try
            {
            //set up to test addBase()
            Entity testEntity = VerySimpleEntity.create("TestEntity");
            Entity baseEntity = VerySimpleEntity.create("BaseEntity");
            Entity proxiedTestEntity = ((EntityProxy) testEntity).getProxiedEntity();

            proxiedTestEntity.addBase(
                    ((EntityProxy) baseEntity).getIdentifier());

            Entity[] basesArray = testEntity.getBases(EntityConstraint.NONE);
            List bases = new ArrayList(basesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < basesArray.length;
                 arrayIndex++)
                {
                bases.add(basesArray[arrayIndex]);
                }

            //assert tests
            Assert.assertTrue(
                    "TestEntity's bases should have contained baseEntity!", 
                    bases.contains(baseEntity));
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testEntityProxyAddBase()
        {
        try
            {
            //set up to test addBase()
            EntityProxy testEntity = (EntityProxy) VerySimpleEntity.create(
                                             "TestEntity");
            EntityProxy baseEntity = (EntityProxy) VerySimpleEntity.create(
                                             "BaseEntity");

            //Entity proxiedTestEntity = ((EntityProxy)testEntity).getProxiedEntity();
            testEntity.addBase(baseEntity);

            Entity[] basesArray = testEntity.getBases(EntityConstraint.NONE);
            List bases = new ArrayList(basesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < basesArray.length;
                 arrayIndex++)
                {
                bases.add(basesArray[arrayIndex]);
                }

            //assert tests
            Assert.assertTrue(
                    "TestEntity's bases should have contained baseEntity!", 
                    bases.contains(baseEntity));
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testRemoveBase()
        {
        try
            {
            //set up to test removeBase()
            Entity testEntity = VerySimpleEntity.create("TestEntity");
            Entity baseEntity = VerySimpleEntity.create("BaseEntity");
            Entity proxiedTestEntity = ((EntityProxy) testEntity).getProxiedEntity();

            proxiedTestEntity.addBase(
                    ((EntityProxy) baseEntity).getIdentifier());
            proxiedTestEntity.removeBase(
                    ((EntityProxy) baseEntity).getIdentifier());

            Entity[] basesArray = testEntity.getBases(EntityConstraint.NONE);
            List bases = new ArrayList(basesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < basesArray.length;
                 arrayIndex++)
                {
                bases.add(basesArray[arrayIndex]);
                }

            //assert tests
            Assert.assertFalse(
                    "TestEntity's bases should not have contained baseEntity!", 
                    bases.contains(baseEntity));
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testEntityProxyRemoveBase()
        {
        try
            {
            //set up to test removeBase()
            EntityProxy testEntity = (EntityProxy) VerySimpleEntity.create(
                                             "TestEntity");
            EntityProxy baseEntity = (EntityProxy) VerySimpleEntity.create(
                                             "BaseEntity");

            //Entity proxiedTestEntity = ((EntityProxy) testEntity).getProxiedEntity();
            testEntity.addBase(baseEntity);
            testEntity.removeBase(baseEntity);

            Entity[] basesArray = testEntity.getBases(EntityConstraint.NONE);
            List bases = new ArrayList(basesArray.length);

            for (int arrayIndex = 0;
                 arrayIndex < basesArray.length;
                 arrayIndex++)
                {
                bases.add(basesArray[arrayIndex]);
                }

            //assert tests
            Assert.assertFalse(
                    "TestEntity's bases should not have contained baseEntity!", 
                    bases.contains(baseEntity));
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testGetAndSetValue()
        {
        try
            {
            Entity testEntity = VerySimpleEntity.create("TestEntity");
            Assert.assertNotSame("value should not be equal to " + 
                                 BooleanEntityValue.TRUE + " yet!", 
                                 BooleanEntityValue.TRUE, testEntity.getValue());
            testEntity.setValue(BooleanEntityValue.TRUE);
            Assert.assertEquals("value should equal " + 
                                BooleanEntityValue.TRUE + "!", 
                                BooleanEntityValue.TRUE, testEntity.getValue());
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testGetAttributes()
        {
        //set up to test getAttributes()
        Entity testEntity;
        Entity attributeEntity;

        try
            {
            testEntity = VerySimpleEntity.create("TestEntity");
            attributeEntity = VerySimpleEntity.create("AttributeEntity");
            testEntity.attach(attributeEntity);

            Entity[] attributesArray = testEntity.getAttributes(
                                               EntityConstraint.NONE);
            Assert.assertEquals(
                    "attributesArray[0] should contain attributeEntity!", 
                    attributeEntity, attributesArray[0]);
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testGetBases()
        {
        //set up to test getBases()
        Entity testEntity;
        Entity baseEntity;

        try
            {
            testEntity = VerySimpleEntity.create("TestEntity");
            baseEntity = VerySimpleEntity.create("BaseEntity");
            ((EntityProxy) testEntity).addBase((EntityProxy) baseEntity);

            Entity[] basesArray = testEntity.getBases(EntityConstraint.NONE);
            Assert.assertEquals("basesArray[0] should contain baseEntity!", 
                                baseEntity, basesArray[0]);
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    public void testIdentifierManager()
        {
        Entity testEntity;
        Entity testEntity2;
        Entity testEntity3;
        Entity testEntity4;

        try
            {
            testEntity = VerySimpleEntity.create("Test");
            testEntity2 = VerySimpleEntity.create("Test");
            testEntity3 = VerySimpleEntity.create("TestYo");
            testEntity4 = VerySimpleEntity.create("TestYo");
            Assert.assertNotSame(
                    "testEntity and testEntity2 should not have the same identifier!", 
                    ((EntityProxy) testEntity).getIdentifier(), 
                    ((EntityProxy) testEntity2).getIdentifier());
            Assert.assertNotSame(
                    "testEntity and testEntity3 should not have the same identifier!", 
                    ((EntityProxy) testEntity).getIdentifier(), 
                    ((EntityProxy) testEntity3).getIdentifier());
            Assert.assertNotSame(
                    "testEntity and testEntity4 should not have the same identifier!", 
                    ((EntityProxy) testEntity).getIdentifier(), 
                    ((EntityProxy) testEntity4).getIdentifier());
            Assert.assertNotSame(
                    "testEntity2 and testEntity3 should not have the same identifier!", 
                    ((EntityProxy) testEntity2).getIdentifier(), 
                    ((EntityProxy) testEntity3).getIdentifier());
            Assert.assertNotSame(
                    "testEntity2 and testEntity4 should not have the same identifier!", 
                    ((EntityProxy) testEntity2).getIdentifier(), 
                    ((EntityProxy) testEntity4).getIdentifier());
            Assert.assertNotSame(
                    "testEntity3 and testEntity4 should not have the same identifier!", 
                    ((EntityProxy) testEntity3).getIdentifier(), 
                    ((EntityProxy) testEntity4).getIdentifier());
            }
        catch (org.datamanager.exception.DataManagerException dataManagerException)
            {
            dataManagerException.printStackTrace();
            }
        }

    private void setSubscribed()
        {
        this.hasSubscribed = true;
        }

    public void testUnsubscribe()
        {
        try
            {
            eventTestEntity = VerySimpleEntity.create("TestEntity");

            DataManagerWorker worker = new DataManagerWorker(new Simpleton()
                {
                private EventHandler eventHandler = new EventHandler()
                    {
                    public void handle(DataManagerEvent event)
                        {
                        if (!(event instanceof SubscribeEventGeneratorEvent))
                            {
                            testString = "Handled!";
                            }
                        }
                    };

                public void process()
                    {
                    try
                        {
                        eventTestEntity.subscribe(eventHandler, 
                                                  EventConstraint.NONE);
                        eventTestEntity.unsubscribe(eventHandler);
                        setSubscribed();
                        }
                    catch (Exception exception)
                        {
                        exception.printStackTrace();
                        fail("Caught an exception (" + exception + 
                                 ") trying to fire our event");
                        }
                    }
                });

            worker.generateEvents(true);
            worker.start();

            // Note we don't want to fire this until we're sure the
            // subscription has taken place.
            DataManagerEvent dataManagerEvent = new DataManagerEvent()
                {
                public EventGenerator getSource()
                    {
                    return eventTestEntity;
                    }

                public long getTime()
                    {
                    return System.currentTimeMillis();
                    }

                public String toString()
                    {
                    return "DataManagerEvent instance";
                    }
                };

            while (!hasSubscribed)
                {
                Thread.currentThread().sleep(1);
                }

            ((EntityProxy) eventTestEntity).fireEvent(dataManagerEvent);

            Thread.currentThread().sleep(1000);

            assertEquals("", testString);
            }
        catch (Exception exception)
            {
            this.fail("Caught an exception (" + exception + 
                      ") trying to fire our event");
            }
        }

    public void testSubscribe()
        {
        try
            {
            eventTestEntity = VerySimpleEntity.create("TestEntity");

            DataManagerWorker worker = new DataManagerWorker(new Simpleton()
                {
                private EventHandler eventHandler = new EventHandler()
                    {
                    public void handle(DataManagerEvent event)
                        {
                        testString = "Handled!";
                        }
                    };

                public void process()
                    {
                    try
                        {
                        eventTestEntity.subscribe(eventHandler, 
                                                  EventConstraint.NONE);
                        setSubscribed();
                        }
                    catch (Exception exception)
                        {
                        exception.printStackTrace();
                        fail("Caught an exception (" + exception + 
                                 ") trying to fire our event");
                        }
                    }
                });

            worker.generateEvents(true);
            worker.start();

            // Note we don't want to fire this until we're sure the
            // subscription has taken place.
            DataManagerEvent dataManagerEvent = new DataManagerEvent()
                {
                public EventGenerator getSource()
                    {
                    return eventTestEntity;
                    }

                public long getTime()
                    {
                    return System.currentTimeMillis();
                    }

                public String toString()
                    {
                    return "DataManagerEvent instance";
                    }
                };

            while (!hasSubscribed)
                {
                Thread.currentThread().sleep(1);
                }

            ((EntityProxy) eventTestEntity).fireEvent(dataManagerEvent);

            Thread.currentThread().sleep(1000);

            assertEquals("Handled!", testString);
            }
        catch (Exception exception)
            {
            this.fail("Caught an exception (" + exception + 
                      ") trying to fire our event");
            }
        }
    }